signature btreedatasig1 =
  sig
    pred isleaf/1 and isnode/1 and mkleaf/1 and mknode/4 and label/2 and
           leftchild/2 and rightchild/2.
  end.

signature btreemem =
  sig
    structure b/sig
    		   pred isleaf/1 and isnode/1 and mkleaf/1 and mknode/4 and
	  		 label/2 and leftchild/2 and rightchild/2.
		end.
    pred newmember/2.
  end.

signature btreeeq =
  sig
    structure c/sig
    		   pred isleaf/1 and isnode/1 and mkleaf/1 and mknode/4 and
	  		 label/2 and leftchild/2 and rightchild/2.
		end.
    pred eqtree/2.
  end.

structure btreedata1/sig
    		         pred isleaf/1 and isnode/1 and mkleaf/1 and mknode/4
			      and label/2 and leftchild/2 and rightchild/2.
		     end =
  struct
    fun leaf/0 and tree/3.          isleaf(leaf).
    isnode(tree(_,_,_)).            mkleaf(leaf).
    mknode(A,L,R,tree(A,L,R)).      label(tree(A,_,_), A).
    leftchild(tree(_,L,_), L).      rightchild(tree(_,_,R), R).
  end.

functor absbtreeeq(x/sig
    		         pred isleaf/1 and isnode/1 and mkleaf/1 and mknode/4
			      and label/2 and leftchild/2 and rightchild/2.
		     end) =
  struct
    structure c = x.
    eqtree(Tree1, Tree2) :- c:isleaf(Tree1), c:isleaf(Tree2).
    eqtree(Tree1, Tree2) :- c:label(Tree1, Label), c:label(Tree2, Label),
          c:leftchild(Tree1, Left1), c:leftchild(Tree2, Left2),
          c:rightchild(Tree1, Right1),c:rightchild(Tree2, Right2),
          eqtree(Left1, Left2), eqtree(Right1, Right2).
  end.

functor absbtreemem(x/sig
    		         pred isleaf/1 and isnode/1 and mkleaf/1 and mknode/4
			      and label/2 and leftchild/2 and rightchild/2.
		     end) =
  struct
    structure b = x.
    newmember(A, Tree) :- b:label(Tree, A).
    newmember(A, Tree) :- b:leftchild(Tree, Left), newmember(A, Left).
    newmember(A, Tree) :- b:rightchild(Tree, Right), newmember(A, Right).
  end.

functor absbtreeutil(x/btreemem, y/btreeeq sharing x:b=y:c) =
  struct
    structure u = x.    structure v = y.
    foobar(Elem, Tree1, Tree2) :- u:newmember(Elem, Tree1),
          v:eqtree(Tree1, Tree2).
  end.

structure btreemem = absbtreemem(btreedata1).
structure btreeeq = absbtreeeq(btreedata1).
structure btreeutil = absbtreeutil(btreemem,btreeeq).

